/*
 * gwinv.h - the GW function and the class used as argument for it
 *
 * The function GW is the user-land function to compute a GW invariant. It
 * takes one argument of type gwlist. A gwlist is just a list of gwentry types.
 * Every gwentry can be either a polynomial (optionally with a multiplicity),
 * in which case it describes a marked point, or an assignment to a special
 * variable of type gwvar (currently the dimension of the variety, the degrees
 * of the variety and the curves, and the genus of the curves). These special
 * variables are defined here, so that the main program doesn't have to
 * define them any more. A typical call to GW looks as follows
 *   GW ( gwlist(N=2) + gwlist (D=1) + gwlist (H*H,2) );
 * which would compute the number of lines through 2 points, assuming that
 * G=0 and E=1 (dimension of the variety). The order of the elements in the
 * gwlist is irrelevant. The special variables are global variables, so they
 * can (but need not) be defined in the GW call. Single statements like e.g.
 *   N=2;
 * outside of a GW function are valid too; they set the default N value for
 * the next GW calls.
 *
 * The gwlist cast can also include a relative multiplicity m and a number
 * n how often the class appears in the invariant:
 *   gwlist (poly, n, m)
 * 
 * In this file, the GW function only checks the input thoroughly, and splits
 * the polynomials into monomials. To actually compute a monomial GW invariant,
 * compute () is called with the invariant converted to type inv.
 *
 */

#ifndef __GW_H__
#define __GW_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "poly.h"
#include "compute.h"



typedef enum {NONE,VARSET,POINT} gwlisttype;



class gwvar { // the class for the special variables:
	      // degree, dimension, genus...
	public:
		char name;
		int value;

		gwvar (char c, int n) {name = c; value = n;};
		gwvar& operator= (const int& n) {
			value = n;
			return *this;
		};
};



class gwentry { // one argument to the GW function
	public:
		gwlisttype type; // what does this entry describe?
		// THE FOLLOWING ONLY RELEVANT IF TYPE IS VARSET.
		char name; // the variable to be set
		int value; // the value
		// THE FOLLOWING ONLY RELEVANT IF TYPE IS POINT.
		poly cls; // the class (polynomial in H and c)
		int mult; // the multiplicity of relative GW theory

		gwentry () {
			type = NONE;
			cls = poly ();
		};
		gwentry (const gwvar& v) {
			type = VARSET;
			name = v.name;
			value = v.value;
			cls = poly ();
		};
		gwentry (const poly& x, int m = 0) {
			type = POINT;
			cls = x;
			mult = m;
		};
		gwentry (const gwentry& x) {
			type = x.type;
			name = x.name;
			value = x.value;
			/*
			 * The following cls statement may look innocent, but
			 * in fact it doubles the cur_ptr of the polynomial,
			 * so that we can do loops over the two copies of
			 * the polynomial independently, although the
			 * polynomial itself is stored in memory only once.
			 *
			 */
			cls = x.cls;
			mult = x.mult;
		};
		gwentry& operator= (const gwentry& x) {
			type = x.type;
			name = x.name;
			value = x.value;
			cls = x.cls;
			mult = x.mult;

			return *this;
		};
};



class gwlist { // the complete list of arguments to the GW function
	public:
		gwlist () {
			alloc = 0;
			data = NULL;
		};
		gwlist (const gwvar& v) {
			alloc = 1;
			assert (data = new gwentry[1]);
			data[0].type = VARSET;
			data[0].name = v.name;
			data[0].value = v.value;
			data[0].cls = poly ();
		};
		gwlist (const poly& x, int m = 0, int c = 1) {
			if (c<=0) {
				alloc = 0;
				data = NULL;
				return;
			}
			alloc = c;
			assert (data = new gwentry[c]);
			for (int i=0; i<c; i++) {
				data[i].type = POINT;
				data[i].cls = x;
				data[i].mult = m;
			}
		};
		gwlist (const gwlist& li) {
			if (!li.alloc) {
				alloc = 0;
				data = NULL;
			};
			alloc = li.alloc;
			assert (data = new gwentry[alloc]);
			for (int i=0; i<alloc; i++)
				data[i] = li.data[i];
		};
		gwlist& operator= (const gwlist& li) {
			if (!li.alloc) {
				alloc = 0;
				data = NULL;
			};
			delete [] data;
			alloc = li.alloc;
			assert (data = new gwentry[alloc]);
			for (int i=0; i<alloc; i++)
				data[i] = li.data[i];
			return *this;
		};
		~gwlist () {delete [] data;};

		// List concatenation
		friend gwlist operator+ (const gwlist&, const gwlist&);

		// The GW function
		friend rational GW (const gwlist&);
	private:
		int alloc;
		gwentry *data;
};



extern gwvar N; // dimension of the variety
extern gwvar D; // degree of the variety
extern gwvar G; // genus of the curves
extern gwvar E; // degree of the curves
extern var H;
extern var psi;
var pc(int); // a primitive cohomology class

#endif

