/*
 * inv.h - a C++ class for the data determining GW invariants
 *
 * This file provides two classes: pdata for the data associated with every
 * marked point (H power, psi power and relative multiplicity), and inv for
 * the data associated with a complete invariant. There are no complicated
 * computations in this file. The H power can be negative, meaning that there
 * is a single primitive cohomology class at this point.
 *
 * Note:
 * - Currently the data fields are designed to describe absolute and relative
 *   invariants in projective spaces and their hypersurfaces.
 * - An invariant is relative if it has at least one point with positive
 *   multiplicity. Note that a relative invariant for a certain degree (E)
 *   and dimension (N) of the variety is in fact an invariant of the ambient
 *   P^(N+1) relative to the degree-E hypersurface! This is important for
 *   computations of the virtual dimension of an invariant.
 *
 */

#ifndef __INV_H__
#define __INV_H__

#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <assert.h>

#include "database.h"



class pdata { // the data associated to every marked point
	public:
		int Hpow; // power of the H class, or a negative value if
			  // this is a primitive cohomology class. In the
			  // latter case, the number does not matter; it just
			  // has the significance that if the same negative
			  // number appears several times in the invariant,
			  // it means that it is the same primitive class.
		int psipow; // power of the psi class
		int mult; // multiplicity for relative invariants

		pdata () {};
		pdata (int i, int j = 0, int k = 0) {
			Hpow = i; psipow = j; mult = k;
		}

		// Comparisons
		friend int cmp (const pdata&, const pdata&);
		friend bool operator== (const pdata& x, const pdata& y)
			{return !cmp (x,y);};
		friend bool operator!= (const pdata& x, const pdata& y)
			{return cmp (x,y);};
		friend bool operator< (const pdata& x, const pdata& y)
			{return cmp (x,y)<0;};
		friend bool operator<= (const pdata& x, const pdata& y)
			{return cmp (x,y)<=0;};
		friend bool operator> (const pdata& x, const pdata& y)
			{return cmp (x,y)>0;};
		friend bool operator>= (const pdata& x, const pdata& y)
			{return cmp (x,y)>=0;};

		// The virtual codimension. Note that we take the _real_
		// codimension as we have to deal with possibly non-algebraic
		// primitive cohomology classes. We have to pass the dimension
		// of the variety to know the codimension of a primitive
		// class.
		int vdim (int n) {
			if (Hpow<0) return n+(2*(psipow+mult-1));
			else return 2*(Hpow+psipow+mult-1);
		};
		// Same as vdim, but ignore multiplicities.
		int vdim_nomult (int n) {
			if (Hpow<0) return n+(2*(psipow-1));
			else return 2*(Hpow+psipow-1);
		};

		// Output
		friend ostream& operator<< (ostream&, const pdata&);
};



class inv {
	public:
		int xdim; // the (complex) dimension of the variety
		int xdeg; // the degree of the variety
		int cgenus; // the genus of the curve
		int cdeg; // the degree of the curve
		int points; // the number of marked points
		pdata *pt; // classes associated to the marked points

		// Constructors and destructors
		inv (int n = 0) {
			points = 0;
			alloc = n;
			pt = NULL;
			if (n) assert (pt = new pdata[n]);
		};
		inv (const inv& x) {
			xdim = x.xdim;
			xdeg = x.xdeg;
			cgenus = x.cgenus;
			cdeg = x.cdeg;
			points = x.points;
			alloc = x.alloc;
			if (alloc) {
				assert (pt = new pdata[alloc]);
			} else pt = NULL;
			if (points) memmove (pt,x.pt,points*sizeof(pdata));
		};
		inv& operator= (const inv& x) {
			xdim = x.xdim;
			xdeg = x.xdeg;
			cgenus = x.cgenus;
			cdeg = x.cdeg;
			points = x.points;
			if (x.points>alloc) {
				alloc = x.points;
				if (pt) delete [] pt;
				if (alloc) assert (pt = new pdata[alloc]);
			}
			if (points) memmove (pt,x.pt,points*sizeof(pdata));
			return *this;
		};
		~inv () {if (pt) delete [] pt;};

		// Point management
		void realloc (int n); // allocate at least n points
		void swap (int m, int n) { // swap two points
			if (m!=n) {
				pdata tmp = pt[m];
				pt[m] = pt[n];
				pt[n] = tmp;
			}
		};
		void sort () {
			// For negative genus invariants, the first two points
			// are never moved
			if (cgenus>=0) internal_sort (-1,-1);
			else internal_sort (2,-1);
		};
		int reduce (); // reduce primitive classes
		void del (int); // delete a marked point
		void delmult () { // delete all multiplicities
			for (int i=0; i<points; i++) pt[i].mult = 0;
		};

		// Output
		friend ostream& operator<< (ostream&, const inv&);

		// Check the invariant
		void check() const;

		// Database functions
		friend void set_key (database&, const inv&);
		friend void set_value (database&, const inv&);
		friend bool get_value (database&, inv&);
	private:
		void internal_sort (int, int); // sort points
		int alloc; // memory allocated for this many number of points
};

#endif

