/*
 * monomial.h - a C++ class for monomial arithmetic
 *
 * This class allows computations with monomials with an arbitrary number
 * of variables (of type var). For example, you can write
 *   var x("x");
 *   var y("y");
 *   monomial m = pow(x,2)*y;
 *   cout << m;
 * which prints x2y. Typecasts from var to monomial are provided, as well as
 * monomial() for the UNDEFINED monomial and monomial(1) for the constant
 * monomial 1. Division of monomials will return UNDEFINED if the result would
 * contain negative exponents. Comparison of monomials is supported using the
 * lexicographical order.
 *
 * Note:
 * - A monomial doesn't include a coefficient. Hence, multiplication with
 *   numbers is not defined. Addition and subtraction are undefined too.
 * - The variables in a monomial are always stored in lexicographical order.
 *   No variable stored in a monomial may occur with exponent 0.
 *
 */

#ifndef __MONOMIAL_H__
#define __MONOMIAL_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "var.h"
#include "database.h"



class monomial_data { // this class not to be used outside this library
	public:
		short vars; // the number of variables

		void monomial_debug (int, int);

		monomial_data (int v = 0) {
			refcnt = 1;
			vars = v;
			if (v) assert (var_exp = new int[v<<1]);
			else var_exp = NULL;
			monomial_debug (refcnt,1);
		};
		~monomial_data () {};

		int& vr (int n) {return var_exp[n<<1];}; // variable number
		int& ex (int n) {return var_exp[(n<<1)+1];}; // exponent

		short operator++ () {
			++refcnt;
			monomial_debug (refcnt,1);
			return refcnt;
		};
		short operator-- () {
			monomial_debug (refcnt-1,-1);
			if (!--refcnt) delete [] var_exp;
			return refcnt;
		};
		short cnt () {return refcnt;};
	private:
		int *var_exp; // an int array with entries:
			      // 0: 1st variable number
			      // 1: 1st variable exponent
			      // 2: 2nd variable number
			      // ...
		short refcnt; // reference counter as in rational, for
			      // memory management of int *var_exp
};



class monomial {
	public:
		monomial () {ptr = NULL;};
		monomial (const monomial& x) {
			ptr = x.ptr;
			if (ptr) ++(*ptr); // raise reference counter
		};
		monomial (int n) { // for the constant 1 monomial
			assert (n==1);
			assert (ptr = new monomial_data);
		};
		monomial (var v) {
			assert (ptr = new monomial_data(1));
			ptr->vars = 1;
			ptr->vr(0) = getnum(v);
			ptr->ex(0) = 1;
		};
		monomial& operator= (const monomial& x) {
			if (this==&x) return *this;
			if (ptr) if ( ! --(*ptr)) delete ptr;
			ptr = x.ptr;
			if (ptr) ++(*ptr);
			return *this;
		};
		~monomial () {
			if (ptr) if ( ! --(*ptr)) delete ptr;
		};

		// Multiplication
                friend monomial operator* (const monomial&, const monomial&);
		monomial& operator*= (const monomial& x)
			{return *this = (*this)*x;};
		// Division
		friend monomial operator/ (const monomial&, const monomial&);
		monomial& operator/= (const monomial& x)
			{return *this = *this/x;};
                // Exponentiation
                friend monomial pow (const monomial&, const int&);

		// Comparison. The basic function returns <0 if the first
		// argument is smaller, >0 if the second is smaller, and 0
		// if they are equal. Comparison is done using the
		// reverse lexicographical order.
		friend int cmp (const monomial&, const monomial&);
		friend bool operator== (const monomial& x, const monomial& y);
		friend bool operator!= (const monomial& x, const monomial& y);
		friend bool operator< (const monomial& x, const monomial& y);
		friend bool operator<= (const monomial& x, const monomial& y);
		friend bool operator> (const monomial& x, const monomial& y);
		friend bool operator>= (const monomial& x, const monomial& y);

		// Casting
		friend bool is_undefined (const monomial& x);
		friend bool is_var (const monomial& x);
		friend var to_var (const monomial& x);
		friend bool is_one (const monomial& x);

		// Degree
		friend int deg (const monomial& x, int varnum);
		friend int deg (const monomial& x, var v);

		// Input and output
		friend ostream& operator<< (ostream&, const monomial&);
		friend istream& operator>> (istream&, monomial&);

		// Database functions
		friend void set_key (database&, const monomial&);
		friend void set_value (database&, const monomial&);
		friend bool get_value (database&, monomial&);

		// Read variables and exponents of a monomial
		int vars () {return (ptr ? ptr->vars : 0);};
		int vr (int n) {return (ptr ? ptr->vr(n) : 0);};
		int ex (int n) {return (ptr ? ptr->ex(n) : 0);};

	private:
		monomial_data *ptr;
};



// Multiplication
monomial operator* (const monomial&, const monomial&);
monomial operator* (const monomial& x, const var& v);
monomial operator* (const var& v, const monomial& x);
monomial operator* (const var& v, const var& w);

// Division
monomial operator/ (const monomial&, const monomial&);
monomial operator/ (const monomial& x, const var& v);
monomial operator/ (const var& v, const monomial& x);
monomial operator/ (const var& v, const var& w);

// Exponentiation
monomial pow (const monomial&, const int&);
monomial pow (const var& v, const int& n);

// Comparison
int cmp (const monomial&, const monomial&);
bool operator== (const monomial& x, const monomial& y);
bool operator!= (const monomial& x, const monomial& y);
bool operator< (const monomial& x, const monomial& y);
bool operator<= (const monomial& x, const monomial& y);
bool operator> (const monomial& x, const monomial& y);
bool operator>= (const monomial& x, const monomial& y);

// Comparisons with variables
bool operator== (const monomial& x, const var& v);
bool operator== (const var& v, const monomial& x);
bool operator!= (const monomial& x, const var& v);
bool operator!= (const var& v, const monomial& x);

// Casting
bool is_undefined (const monomial& x);
bool is_var (const monomial& x);
var to_var (const monomial& x);
bool is_one (const monomial& x);

// Degree
int deg (const monomial& x, int varnum);
int deg (const monomial& x, var v);

#endif

