/*
 * poly.h - a C++ class for polynomial arithmetic
 *
 * This is a class for computations with polynomials with rational coefficients
 * in an arbitrary number of variables. For example, you can write
 *   var x("x");
 *   var y("y");
 *   poly p = pow(x-y,3);
 *   cout << subs (p,y,1);
 * which prints x3-3x2+3x-1. Usage and implementation are very similar to
 * the rational and monomial classes.
 *
 * Note:
 * - A polynomial can be UNDEFINED. This happens e.g. if a division "doesn't
 *   go". So, for example, (x2+2x+1)/(x+1) yields the correct result x+1,
 *   but (x2+2x+1)/(x+2) gives UNDEFINED.
 * - Internally, a polynomial is a list of pairs (coefficient, monomial).
 *   This list is always in sorted order, and none of the coefficients is 0.
 *   (If the polynomial is 0, the list is empty). At the beginning of the
 *   list there is a "dummy element" whose coefficient and monomial are
 *   ignored (and are always UNDEFINED). This first element contains the
 *   reference counter for the memory management. The reference counters
 *   in the other list elements are ignored.
 * - There is no explicit function to sort the summands in a polynomial.
 *   All operations will produce their result in sorted order already.
 * - Typecasts to poly are provided, and usually need not be specified
 *   if they go "upwards". So, e.g. the expression p+4 is valid if p is
 *   a polynomial (the int 4 is automatically upgraded to a poly as well).
 *   A backward typecast works like this: to_int(p) for p a poly gives the
 *   corresponding int. This is well-defined only if p actually fits in an int
 *   (i.e. is a constant and is small enough), which can be tested with is_int.
 * - Comparisions: Polynomials can only be tested for equality.
 * - Some of the functions are optimized if one of the arguments is easy:
 *   e.g. multiplication with an int is done differently from multiplication
 *   of two polynomials. The functions +=, -=, *=, /= have special
 *   implementations that try to avoid frequent re-allocation of memory.
 *   As in the other mathematical classes, the use of reference counters make
 *   assignments very fast (they don't result in a deep copy of the poly).
 * - Polynomials can be read sequentially. For example, if p is a poly,
 *     p.mon_first ();
 *     while (p.mon_next ()) cout << p.mon_coeff << " ";
 *   will print all the (rational) coefficients of the polynomial. In the same
 *   way, mon_vars gives the monomials. mon_first _has_ to be called before
 *   the elements can be read. Any assignment to p will invalidate the pointer,
 *   so that mon_first has to be called again before the elements can be read
 *   with mon_next, mon_coeff, and mon_vars.
 *
 */

#ifndef __POLY_H__
#define __POLY_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "rational.h"
#include "monomial.h"
#include "database.h"



class poly;

		// Addition and subtraction. There is only one actual
		// function; the rest is done via typecasting.
		// Multiplication and division with "scalar" types
		// poly pmul (const poly&,const monomial&,const rational&);
		// poly pdiv (const poly&,const monomial&,const rational&);
		// Exponentiation
		// poly pow (const poly&, const int&);

		// Signs
		
		// Comparison. Only equality testing is possible
		// int pcmp (const poly&, const poly&);
		
		// Conversion for division
		
		// Degrees and coefficients
		//int deg (const poly&, int);
		// Substitution
		//poly subs (const poly&, const var&, const poly&);

		
		// Input and output
		//ostream& operator<< (ostream&, const poly&);
		//istream& operator>> (istream&, poly&);

		// Database functions
		//void set_key (database&, const poly&);
		//void set_value (database&, const poly&);
		//bool get_value (database&, poly&);





class poly_data { // this class not to be used outside this library
	public:
		rational cof;
		monomial mon;
		poly_data *next;

		void poly_debug (int);

		poly_data () {
			refcnt = 1;
			poly_debug (1);
			next = NULL;
		};
		~poly_data () {
			poly_debug (-1);
		};

		// ++ and -- may only be applied to the first poly_data of a
		// polynomial, where the reference counter is relevant.
		short operator++ () {
			++refcnt;
			return refcnt;
		};
		short operator-- () {
			if (!--refcnt) {
				// Delete all attached poly_data
				poly_data *p;
				while (next) {
					p = next->next;
					delete next;
					next = p;
				};
			};
			assert (refcnt>=0);
			return refcnt;
		};
		short cnt () {return refcnt;};
	private:
		short refcnt; // reference counter for memory management,
			      // as in rational. Relevant only in the first
			      // poly_data of a polynomial.
};



class poly {
	public:
		// Constructors and destructors
		poly () {ptr = NULL;};
		poly (const poly& x) {
			ptr = x.ptr;
			if (ptr) ++(*ptr); // raise reference count
		};
		poly (const monomial& x) {
			if (is_undefined (x)) ptr = NULL;
			else {
				assert (ptr = new poly_data);
				assert (ptr->next = new poly_data);
				ptr->next->cof = 1;
				ptr->next->mon = x;
			}
		};
		poly (const var& x) {
			assert (ptr = new poly_data);
			assert (ptr->next = new poly_data);
			ptr->next->cof = 1;
			ptr->next->mon = monomial (x);
		};
		poly (const rational& x) {
			if (is_undefined (x)) ptr = NULL;
			else {
				assert (ptr = new poly_data);
				if (x!=0) {
					assert (ptr->next = new poly_data);
					ptr->next->cof = x;
					ptr->next->mon = 1;
				}
			}
		};
		poly (const int& x) {
			assert (ptr = new poly_data);
			if (x) {
				assert (ptr->next = new poly_data);
				ptr->next->cof = rational (x);
				ptr->next->mon = 1;
			}
		};
		poly& operator= (const poly& x) {
			if (this==&x) return *this;
			if (ptr) if ( ! --(*ptr)) delete ptr;
			ptr = x.ptr;
			if (ptr) ++(*ptr);
			return *this;
		};
		~poly () {
			if (ptr) if ( ! --(*ptr)) delete ptr;
		};

		// Addition and subtraction. There is only one actual
		// function; the rest is done via typecasting.
		friend poly paddsub (const poly&, const poly&, int);

		// Addition and subtraction in =+ / -= version. We provide
		// optimized functions for this.
		poly& paddsub2 (const poly&, int);
		#define PADDSUB2(TYPE)					\
		poly& operator+= (const TYPE& x)			\
			{return *this = paddsub2 (x,0);};		\
		poly& operator-= (const TYPE& x)			\
			{return *this = paddsub2 (x,1);};
		PADDSUB2(poly);
		PADDSUB2(monomial);
		PADDSUB2(var);
		PADDSUB2(rational);
		PADDSUB2(int);

		// Multiplication and division with "scalar" types
		friend poly pmul (const poly&,const monomial&,const rational&);
		friend poly pdiv (const poly&,const monomial&,const rational&);

		// "Scalar" multiplication and division in *= / /= form. We
		// provide optimized functions for this.
		poly& pmul2 (const monomial&, const rational&);
		poly& pdiv2 (const monomial&, const rational&);
		poly& operator*= (const monomial& x)
			{return *this = pmul2 (x,rational ());};
		poly& operator*= (const var& x)
			{return *this = pmul2 (x,rational ());};
		poly& operator*= (const rational& x)
			{return *this = pmul2 (monomial(),x);};
		poly& operator*= (const int& x)
			{return *this = pmul2 (monomial(),x);};
		poly& operator/= (const monomial& x)
			{return *this = pdiv2 (x,rational ());};
		poly& operator/= (const var& x)
			{return *this = pdiv2 (x,rational ());};
		poly& operator/= (const rational& x)
			{return *this = pdiv2 (monomial(),x);};
		poly& operator/= (const int& x)
			{return *this = pdiv2 (monomial(),x);};

		// Multiplication and division of polynomials
		friend poly operator* (const poly&, const poly&);
		poly& operator*= (const poly& x)
			{return *this = (*this)*x;};
		friend poly operator/ (const poly&, const poly&);
		poly& operator/= (const poly& x)
			{return *this = (*this)/x;};

		// Exponentiation
        	friend poly pow (const poly&, const int&);
	
		// Comparison. Only equality testing is possible
		friend int pcmp (const poly&, const poly&);

		// Casting
		friend bool is_undefined (const poly& x);
		friend bool is_zero (const poly& x);
		friend bool is_monomial (const poly& x);
		friend monomial to_monomial (const poly& x);
		friend bool is_var (const poly& x);
		friend var to_var (const poly& x);
		friend bool is_rational (const poly& x);
		friend rational to_rational (const poly& x);
		friend bool is_int (const poly& x);
		friend int to_int (const poly& x);

		// Degrees and coefficients
		friend int deg (const poly&, int);
		friend poly coeff (const poly&, int, int);

		// Substitution
		friend poly subs (const poly&, const var&, const poly&);

		// Sequential reading of polynomials
		void mon_first () {cur_ptr = ptr;};
		bool mon_next () {
			if (cur_ptr) cur_ptr = cur_ptr->next;
			return cur_ptr;
		};
		rational mon_coeff () {
			if (!cur_ptr) return rational ();
			return cur_ptr->cof;
		};
		monomial mon_vars () {
			if (!cur_ptr) return monomial ();
			return cur_ptr->mon;
		};

		// Input and output
		friend ostream& operator<< (ostream&, const poly&);
		friend istream& operator>> (istream&, poly&);

		// Database functions
		friend void set_key (database&, const poly&);
		friend void set_value (database&, const poly&);
		friend bool get_value (database&, poly&);
	private:
		poly_data *ptr;
		poly_data *cur_ptr; // pointer for sequential reading
};



// Addition and subtraction
poly paddsub (const poly&, const poly&, int);
#define PPADDSUB(TYPE1,TYPE2)			        \
poly operator+ (const TYPE1& x, const TYPE2& y);	\
poly operator- (const TYPE1& x, const TYPE2& y);
PPADDSUB (poly,poly);
PPADDSUB (poly,monomial);
PPADDSUB (monomial,poly);
PPADDSUB (monomial,monomial);
PPADDSUB (poly,var);
PPADDSUB (var,poly);
PPADDSUB (var,var);
PPADDSUB (monomial,var);
PPADDSUB (var,monomial);
PPADDSUB (poly,rational);
PPADDSUB (monomial,rational);
PPADDSUB (var,rational);
PPADDSUB (rational,poly);
PPADDSUB (rational,monomial);
PPADDSUB (rational,var);
PPADDSUB (poly,int);
PPADDSUB (monomial,int);
PPADDSUB (var,int);
PPADDSUB (int,poly);
PPADDSUB (int,monomial);
PPADDSUB (int,var);

// Multiplication and division with "scalar" types
poly pmul (const poly&,const monomial&,const rational&);
poly pdiv (const poly&,const monomial&,const rational&);
#define PPMULDIVM(TYPE1,TYPE2)				\
poly operator* (const TYPE1& x, const TYPE2& y);	\
poly operator/ (const TYPE1& x, const TYPE2& y);	\
poly operator* (const TYPE2& y, const TYPE1& x);
#define PPMULDIVR(TYPE1,TYPE2)				\
poly operator* (const TYPE1& x, const TYPE2& y);	\
poly operator/ (const TYPE1& x, const TYPE2& y);	\
poly operator* (const TYPE2& y, const TYPE1& x);
PPMULDIVM (poly,monomial);
PPMULDIVM (poly,var);
PPMULDIVR (poly,rational);
PPMULDIVR (poly,int);
PPMULDIVR (monomial,rational);
PPMULDIVR (monomial,int);
PPMULDIVR (var,rational);
PPMULDIVR (var,int);

// Multiplication and division of polynomials
poly operator* (const poly&, const poly&);
poly operator/ (const poly&, const poly&);

// Exponentiation
poly pow (const poly&, const int&);

// Signs
poly operator+ (const poly& x);
poly operator+ (const monomial& x);
poly operator+ (const var& x);
poly operator- (const poly& x);
poly operator- (const monomial& x);
poly operator- (const var& x);

// Comparison
int pcmp (const poly&, const poly&);
bool operator== (const poly& x, const poly& y);
bool operator!= (const poly& x, const poly& y);
#define PPCMP(TYPE,COMP,CONV)			\
bool operator== (const poly& x, const TYPE& y); \
bool operator== (const TYPE& y, const poly& x); \
bool operator!= (const poly& x, const TYPE& y); \
bool operator!= (const TYPE& y, const poly& x);
PPCMP (monomial,is_monomial,to_monomial);
PPCMP (var,is_var,to_var);
PPCMP (rational,is_rational,to_rational);
PPCMP (int,is_int,to_int);

// Casting
bool is_undefined (const poly& x);
bool is_zero (const poly& x);
bool is_monomial (const poly& x);
monomial to_monomial (const poly& x);
bool is_var (const poly& x);
var to_var (const poly& x);
bool is_rational (const poly& x);
rational to_rational (const poly& x);
bool is_int (const poly& x);
int to_int (const poly& x);

// Conversion for division
poly divconv (var x);
poly divconv (monomial x);
poly divconv (poly x);

// Degrees and coefficients
int deg (const poly&, int);
int deg (const poly& x, const var& v);
poly coeff (const poly&, int, int);
poly coeff (const poly& x, const var& v, int d);
poly coeff (const monomial& x, const var& v, const int& d);

// Substitution
poly subs (const poly&, const var&, const poly&);

// Database functions
void set_key (database&, const poly&);
void set_value (database&, const poly&);
bool get_value (database&, poly&);

#endif

