/*
 * rational.h - a class for rational number arithmetic
 *
 * This is a C++ wrapper library for rational number arithmetic using the
 * GMP library (version >= 4). It behaves exactly as expected, for example
 * you can do the following:
 *
 *   rational x,y,z;
 *
 *   x = 2; y = 3;
 *   z = x/y;
 *   cout << z;
 *
 * (which prints 2/3), or
 *
 *   for (rational x=rational(1)/2; x<=5; x+=1) cout << x << " ";
 *
 * which prints 1/2 3/2 5/2 7/2 9/2. All arithmetic operations and
 * comparisons are defined, as well as functions to check if a number is an
 * integer or fits in an int. See the class definition below. There is no
 * limit on the size of the numbers.
 *
 * Note:
 * - The final program has to be linked with -lgmp.
 * - A rational number can be UNDEFINED. For example, a declaration
 *   "rational x;" leaves x undefined. Division by zero (and similar things)
 *   also gives a result of UNDEFINED (rather than causing the program to
 *   stop with an error). Any operation with an undefined number yields again
 *   UNDEFINED. The functions set_undefined and is_undefined can set and check
 *   this state.
 * - The functions are "smart" in their memory management in the sense that
 *   deep copies of rationals are avoided wherever possible. For example, the
 *   statement
 *     z = x+y;
 *   (with x, y, z of type rational) will result in only one call of mpq_add,
 *   rather than one call of mpq_add plus an additional mpq_set from a
 *   temporary variable to the final destination z. Similarly, the statement
 *     z = z+x;
 *   does not need more GMP calls than
 *     z += x;
 *   This is achieved by introducing an additional class rational_data, to
 *   which rational is a pointer. The class rational_data is not intended to
 *   be used outside this library. It is possible that more than one rational
 *   variable points to the same rational_data. For example, a statement
 *     z = x;
 *   doesn't copy any rational number, but just makes z and x point to the
 *   same memory address. An element of rational_data remembers how many
 *   variables point to it, and the mpq variable is cleared only if this
 *   number drops to zero. Of course, a statement like
 *     z += x;
 *   then has to check first if z points to the same rational_data as some
 *   other variable. If so, the result has to be stored at a different
 *   address, and z then has to point to this new address. All this
 *   "reference counter" management is done in the class rational_data.
 * - The function divconv converts ints and rationals to rationals. So by
 *   replacing x/y by divconv(x)/divconv(y) in a C++ source text we'll always
 *   compute '/' fractions over the rationals.
 * - The doubly-rational operators (+,-,...,=,!=,...) are defined multiple
 *   times, for (rational, rational), (rational, int), and (int, rational).
 *   This is to provide exact type matches when ints and rationals are
 *   intermixed. If we did't do this, the compiler would get confused when
 *   we introduce polynomials, and an addition of type (rational,int) could
 *   be typecast both to (rational,rational) and (poly,poly). We don't define
 *   the operators for integral types other than int, so the application
 *   program should always use int for computations intermixed with rational
 *   numbers that require "short" integers.
 * - Database functions set_key, set_value, get_value are provided for rational
 *   numbers (see database.h). This allows for memory-efficient storing and
 *   reading of rational numbers in databases. WARNING: Reading rational
 *   numbers from a database is done using some undocumented gmp functions
 *   and internal data types, as there are no official functions to provide
 *   the same functionality at the same speed.
 * - We provide a fast factorial function that stores its values in an array.
 *
 */

#ifndef __RATIONAL_H__
#define __RATIONAL_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>
#include <gmp.h>

#include "database.h"



class rational_data { // this class not to be used outside this library
	public:
		mpq_t number;

		void rational_debug (mpq_t, int, int);

		rational_data () {
			refcnt = 1; mpq_init (number);
			rational_debug (number, refcnt, 1);
		};
		~rational_data () {};

		short operator++ () {
			++refcnt;
			rational_debug (number, refcnt, 1);
			return refcnt;
		};
		short operator-- () {
			rational_debug (number, refcnt-1, -1);
			if (!--refcnt) mpq_clear (number);
			return refcnt;
		};
		short cnt () {return refcnt;};
	private:
		short refcnt; // reference counter
};



class rational {
	public:
		rational () {ptr = NULL;};
		rational (const rational& x) {
			ptr = x.ptr;
			if (ptr) ++(*ptr); // raise reference count
		};
		rational (int num, int den=1) { /* we MUST use the canonical
						   form for this constructor */
			assert (ptr = new rational_data);
			mpq_set_si (ptr->number,num,den);
		};
		rational (char *str) {
			assert (ptr = new rational_data);
			mpq_set_str (ptr->number,str,10);
		};
		rational& operator= (const rational& x) {
			if (this==&x) return *this;
			if (ptr) if ( ! --(*ptr)) delete ptr;
			ptr = x.ptr;
			if (ptr) ++(*ptr);
			return *this;
		};
		~rational () {
			if (ptr) if ( ! --(*ptr)) delete ptr;
		};

                // Arithmetic operations
		friend rational operator+ (const rational&, const rational&);
		friend rational operator- (const rational&, const rational&);
		friend rational operator* (const rational&, const rational&);
		friend rational operator/ (const rational&, const rational&);
		friend rational operator<< (const rational &, const unsigned&);
		friend rational operator>> (const rational &, const unsigned&);
		friend rational pow (const rational&, const int&);
		rational& operator+= (const rational&);
		rational& operator-= (const rational&);
		rational& operator*= (const rational&);
		rational& operator/= (const rational&);
		rational& operator<<= (const unsigned&);
		rational& operator>>= (const unsigned&);
		friend rational operator- (const rational&);

                // Comparisons
		friend bool operator== (const rational&, const rational&);
		friend bool operator!= (const rational&, const rational&);
		friend bool operator< (const rational&, const rational&);
		friend bool operator<= (const rational&, const rational&);
		friend bool operator> (const rational&, const rational&);
		friend bool operator>= (const rational&, const rational&);

		// Casting
		friend bool is_undefined (const rational& x);
		// is it an integer (i.e. denominator 1) ?
		friend bool is_integer (const rational&);
		// is it a machine int?
		friend bool is_int (const rational&);
		friend int to_int (const rational& x);
		friend bool is_zero (const rational& x);

		friend rational divconv (int n);
		friend rational divconv (rational& x);

		friend ostream& operator<< (ostream&, const rational&);
		friend istream& operator>> (istream&, rational&);

		// Fast factorial function
		friend rational factorial (int);

		// Database functions
		friend void set_key (database&, const rational&);
		friend void set_value (database&, const rational&);
		friend bool get_value (database&, rational&);
	private:
		rational_data *ptr;
};



// Arithmetic operations
rational operator+ (const rational&, const rational&);
rational operator+ (const rational& x, const int& y);
rational operator+ (const int& x, const rational& y);
rational operator- (const rational&, const rational&);
rational operator- (const rational& x, const int& y);
rational operator- (const int& x, const rational& y);
rational operator* (const rational&, const rational&);
rational operator* (const rational& x, const int& y);
rational operator* (const int& x, const rational& y);
rational operator/ (const rational&, const rational&);
rational operator/ (const rational& x, const int& y);
rational operator/ (const int& x, const rational& y);
rational operator<< (const rational &, const unsigned&);
rational operator>> (const rational &, const unsigned&);
rational pow (const rational&, const int&);
rational pow (const int& x, const int& n);
rational operator+ (const rational& x);
rational operator- (const rational&);

// Comparisons
bool operator== (const rational&, const rational&);
bool operator== (const rational& x, const int& y);
bool operator== (const int& x, const rational& y);
bool operator!= (const rational&, const rational&);
bool operator!= (const rational& x, const int& y);
bool operator!= (const int& x, const rational& y);
bool operator< (const rational&, const rational&);
bool operator< (const rational& x, const int& y);
bool operator< (const int& x, const rational& y);
bool operator<= (const rational&, const rational&);
bool operator<= (const rational& x, const int& y);
bool operator<= (const int& x, const rational& y);
bool operator> (const rational&, const rational&);
bool operator> (const rational& x, const int& y);
bool operator> (const int& x, const rational& y);
bool operator>= (const rational&, const rational&);
bool operator>= (const rational& x, const int& y);
bool operator>= (const int& x, const rational& y);

// Casting
bool is_undefined (const rational& x);
// is it an integer (i.e. denominator 1) ?
bool is_integer (const rational&);
// is it a machine int?
bool is_int (const rational&);
int to_int (const rational& x);
bool is_zero (const rational& x);

rational divconv (int n);
rational divconv (rational& x);
rational factorial (int);

// Database functions
void set_key (database&, const rational&);
void set_value (database&, const rational&);
bool get_value (database&, rational&);

#endif

