/*
 * relinv.h - compute GW splittings in relative theory
 *
 * This C++ class contains the main procedures to compute relative GW
 * invariants using the splitting formula of relative GW theory. In short,
 * if you want to compute an invariant iv using relative GW splitting, write
 *   relinv rv (iv,lvl,skip);
 *
 *   do {
 *     ...
 *   } while rv.next();
 * This will loop over all splittings of the invariant iv onto any number of
 * components. The number lvl is the level of the computation and only used
 * for debugging output. As demanded by relative GW theory, the following
 * splittings are performed:
 * - The degree of the curve is split onto any number of components. Component
 *   number 0 (the 'internal' component) is special in that it can have
 *   degree 0; the others must have positive degree.
 * - The genus of the curve is split onto all components in every possible way.
 * - All classes in the original invariant iv are split in any possible way
 *   onto the components, except for the class with index skip that is always
 *   put on component 0. The number skip can be left out. The classes that
 *   are put onto component 0 lose their multiplicities, making the invariant
 *   into an absolute one. We can handle primitive classes as well; they
 *   will be included in the diagonal split whenever needed.
 * - Every component i>0 is attached to component 0 at n points, where n can
 *   vary. This contributes n-1 to the total genus of the curve. At each of
 *   these points, we do a diagonal splitting and attach any multiplicity.
 * There are some restrictions though:
 * - The components i>0 must have virtual dimension zero.
 * - The total multiplicity of the external components must not exceed the
 *   maximum value cdeg*xdeg.
 * - Primitive classes may only occur in pairs in the external invariants.
 *
 * In the above do-while loop, the current splitting can be read as follows:
 * - rv.cpts () is the number of external components.
 * - rv.cpt (iv,n) sets the invariant iv (that must have enough memory
 *   allocated) to invariant number n of the splitting.
 * - rv.cof () is the coefficient with which the invariant occurs. This
 *   includes combinatorial factors as well as the multiplicity factors from
 *   relative GW theory.
 * - rv.mult () returns 0 if and only if the "relative GW multiplicity
 *   condition" is satisfied, i.e. if the sum of the internal multiplicities
 *   is equal to the sum of the external multiplicities plus the intersection
 *   product of C_0 with the hypersurface.
 *
 * The next() function is smart in the sense that splittings that differ only
 * by a symmetry are returned only once (with a corresponding symmetry factor
 * included in rv.cof() ). This includes
 * - symmetries among the components,
 * - symmetries among the nodes where the external components join the
 *   internal one,
 * - symmetries among the primitive classes.
 * It should be quite good at only returning those configurations that give
 * rise to non-zero contributions, though this is not guaranteed.
 *
 * The fact that next() is quite smart comes at the price that its
 * implementation is pretty complicated. At every stage do we have to keep
 * track which components are currently identical.
 *
 * For details of the implementation, see below...
 *
 */

#ifndef __RELINV_H__
#define __RELINV_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "rational.h"
#include "inv.h"



class relinv { // class for relative splittings
	public:
		relinv (inv&, int, int = -1);
		~relinv () {
			delete [] value_ptr;
			delete [] same_ptr;
			delete [] node_ptr;
			delete [] cls_vdim;
			delete [] cpt_vdim;
			delete [] cpt_tm;
			delete [] cpt_prim;
			delete [] cls;
		};

		bool next (); // Go to the next splitting, return 0 if there
			      // are no more splittings.
		int cpts () {return ncpt-1;}; // The number of external
					     // components.
		void cpt (inv&, int); // get a certain component
		rational cof (); // get the coefficient
		int mult (); // is the multiplicity condition OK?

		// Print a relinv (mainly for debugging)
		friend ostream& operator<< (ostream&, relinv&);

	private:
		// Some values are copied directly from iv:
		int points; // = iv.points
		int totalgenus; // = iv.cgenus
		int xdim; // = iv.xdim
		int xdeg; // = iv.xdeg

		/*
		 * For better performance, classes in relinv are stored
		 * in a different way than in inv. If a class occurs multiple
		 * times, we don't write it into the invariant that often, but
		 * rather keep a list of all classes occuring, and store for
		 * every component how often it contains it. Therefore we need
		 * to have a list of all _distinct_ classes in iv. This is
		 * stored here:
		 *
		 */
		int ncls; // The number of different classes (plus 3, because
			  // counting starts at index 3, see below).
		pdata *cls; // The classes themselves. They are only set
			    // by the relinv constructor and never changed
			    // later.

		int ncpt; // The number of components in the current splitting.
		int lvl; // The level of the computation.
		int skip; // Is there a class that has to be put on cpt 0?
		pdata skipclass; // the class that is to be put on cpt 0?
		int noclass; // 1 if there are no moving classes. There are
			     // faster algorithms in this case for some
			     // functions.

		// Note that we take care of primitive classes as well. Here
		// we store how many such classes there are. It's only set
		// by the constructor and never changed afterwards.
		rational prims;

		// The multiplicities coming from multinomial factors of class
		// splittings. This is constantly updated whenever the class
		// splitting changes.
		rational coeff;

		// The minimum H power to put on an external component. This
		// is 0 unless there are no classes to split and the genus
		// is 0, in which case we set it to a higher value to avoid
		// some unnecessary looping.
		int node_minclass;

		// Now the various arrays...

		// The virtual dimensions of the classes to split. Entries
		// are in 1:1 correspondence with cls.
		int *cls_vdim;

		// The virtual dimensions of the components. Entries are
		// labeled by component number (>0) and are updated whenever
		// degree, genus or class splittings change.
		int *cpt_vdim;

		// The total relative multiplicity of the components. Entries
		// are labeled by component number (>0) and are updated
		// whenever the class splittings change.
		int *cpt_tm;

		/*
		 * This one is a little tricky. Recall that we do handle
		 * primitive classes as well, and we want to do it in a smart
		 * way, i.e. we want to use the symmetry that all primitive
		 * classes give the same invariant. However, if an invariant
		 * contains more than 1 primitive class, it makes a difference
		 * whether these classes are the same or not. So if there is
		 * no primitive class yet anywhere in the invariants, and we
		 * have to add one for diagonal splittings, we can just add
		 * the primitive class -1 (recall that primitive classes are
		 * recognized by negative H powers) and multiply the resulting
		 * number with the number of primitive classes. If there is
		 * a primitive class -1 already there however, we have to
		 * admit both -1 (same one) as well as -2 (different one).
		 * If we have to add several diagonal classes, this will go on.
		 *
		 * In the following array, we store:
		 * - cpt_prim[0] is the last primitive class occurring in the
		 *   invariant iv that we split (a negative number that denotes
		 *   the class, or 0 if there are no primitive classes).
		 * - cpt_prim[i] for i>0 is the last primitive class occuring
		 *   either in iv, or in diagonal nodes from components >i.
		 *
		 */
		int *cpt_prim;

		/*
		 * The following two arrays are 2-dimensional. The first
		 * index is the number of the component, and the second one
		 * is the number of the class (in 1:1 correspondence with
		 * cls). The first three class indices are special:
		 * - 0: degree of the component
		 * - 1: genus plus (connecting diagonal nodes minus 1) - this
		 *      is the total genus contribution of this component.
		 * - 2: the number of connecting diagonal nodes minus 1, i.e.
		 *      the number of loops at the component.
		 * The array at value_ptr contains the actual values. The array
		 * at same_ptr contains the so-called 'same values' that are
		 * used to keep track of which components are identical. More
		 * precisely, same(m,n) stores values which components are
		 * identical if we only consider degree, genus, loops, and
		 * classes up to n (indexed as in cls). Special values are:
		 * - n=0: degree only
		 * - n=1: degree and (genus+loops)
		 * - n=2: degree, genus, and loops
		 * - n=ncls: everything as above plus node classes
		 * - n=ncls+1: everything as above plus node classes and
		 *   multiplicities
		 * So, the values at n=ncls+1 indicate whether components
		 * are 'really' identical after putting all the structure on
		 * them. For a fixed row (fixed n), the same(m,.) values
		 * mean the following: same(m,i) is the number of components,
		 * starting from i, that are identical (they are then always
		 * adjacent to each other). These values are meaningful only
		 * at the _start_ of a block of same classes. For example,
		 * if components 1,2,3 as well as 5,6 are identical, the 7
		 * same values (for components 0 to 6) are
		 *   1,3,*,*,1,2,*
		 * where * stands for an undefined value.
		 * The general philosophy is that two components are the same
		 * on some 'level' n if and only if they are the same on
		 * the previous level n-1, and the data put on in level n is
		 * the same as well.
		 *
		 */
		int *value_ptr;
		int *same_ptr;
		int firstmax; // factor to make 1-dimensional array into
			      // 2-dimensional one.
		// The following functions are just for convenient access
		// to the arrays.
		int& value(int m, int n) {return value_ptr[n*firstmax+m];};
		int& same(int m, int n) {return same_ptr[n*firstmax+m];};
		int& cdeg (int n) {return value(n,0);};
		int& cgenus (int n) {return value(n,1);};
		int& cloop (int n) {return value(n,2);};

		/*
		 * Another 2-dimensional array that stores the node
		 * configurations (in fact, it is three arrays, unified in
		 * one). In all three, the first index is the component number
		 * (at least 1), and the second one is the node number
		 * (running from 0 to cloop(component)).
		 * - node_class: The class (H power, or negative value if
		 *   primitive class) at the node.
		 * - node_mult: The relative multiplicity at the node.
		 * - node_same: A 'same value' as above in same(m,n), but
		 *   for nodes within a component, and considering _only_
		 *   the classes at the nodes (not the multiplicities).
		 *
		 */
		int *node_ptr;
		int& node_mult(int m, int n) {
			return node_ptr[3*n*firstmax+m];
		};
		int& node_class(int m, int n) {
			return node_ptr[(3*n+1)*firstmax+m];
		};
		int& node_same(int m, int n) {
			return node_ptr[(3*n+2)*firstmax+m];
		};

		// Here are various internal functions that advance the
		// current splitting in some respect (e.g. advance degree
		// splitting). See relinv.cc for more details.
		bool deg_next ();
		bool same_next (int, int);
		bool group_next (int);
		bool loop_next ();
		bool nodemult_init (int);
		bool nodemult_next (int);
		bool nodeclass_init (int);
		bool nodeclass_next (int);
		bool node_init ();
		bool node_next ();
};

#endif

