/*
 * split2.h - functions for computing invariants by splitting them into
 *            two components
 *
 * This is for absolute invariants only! Also, we cannot deal with primitive
 * classes currently. What we would have to change to support primitive
 * classes:
 * - change the virtual dimension count to real dimensions
 * - in the diagonal splitting, allow primitive classes in the diagonal. This
 *   is quite tricky because we have to sum over _all_ primitive classes,
 *   so we have to check first which primitive classes already occur, and
 *   then sum over those diagonal classes separately.
 *
 * Note: Because of primitive class complications, elliptic TRR (with a loop
 * term) is currently not supported if the degree of the variety is bigger
 * than 1.
 *
 */

#ifndef __SPLIT2_H__
#define __SPLIT2_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "gwinv.h"
#include "database.h"
#include "compute.h"

// An integer division function that really rounds down.

int idiv (int, int);

// Compute all splittings of cohomology classes onto 2 components. See
// split2.cc for details.

bool next_pair (inv **, int, int, int&, rational&);

/* 
 * Compute a sum of invariants corresponding to 2-component splittings. The
 * first fv and fw classes of the invariants iv and iw are fixed (e.g. fv=fw=3
 * for the WDVV equations). When calling split2_sum, iv and iw must have enough
 * space allocated to accomodate the maximal number of classes that can occur
 * on them. All non-fixed classes must be in iv (i.e. we must have fw =
 * iw.points). The first class iv.pt[0] and iw.pt[0] must be left empty; it
 * will be filled by the diagonal class. The only data of iw that has to be
 * filled in is the genus and the fixed classes; the rest will be filled from
 * iv. We sum over all class splittings, degree splittings, and diagonal sums.
 * The genus is fixed on both sides. The first skip terms are left out. This
 * is probably only useful for skip=1 if the first term (i.e. the invariant iv
 * as it is passed, with maximal degree) is the invariant that is to be
 * computed. The relevant parts of iv and iw will remain unchanged by the
 * function.
 *
 */

rational split2_sum (inv& iv, inv& iw, int fv, int fw, int skip, int lvl);

/*
 * Compute an invariant by splitting it into two components, if possible.
 * This includes WDVV and (some) topological recursion relations. The classes
 * MUST be sorted (we check if there are any descendants by just looking at
 * the last class).
 *
 */

rational split2 (inv&, int);

#endif

