/*
 * var.h - a class for formal (mathematical) variables
 *
 * This very simple class provides the framework to define C++ variables
 * as "mathematical variables". Such a variable is described by a name (an
 * upper or lower case letter; case is significant) and optionally an index
 * (a non-negative number). For example, the declaration
 *   var x('x');
 * defines the C++ variable x to be the mathematical variable x. To define
 * a variable with an index, write
 *   var x('x',3);
 * which generates the mathematical variable x[3]. x[0] is equivalent to x
 * (no index). There are also "unnamed variables" that are generated by
 *   var x;
 * Every time a new var is created without specifying a mathematical name for
 * it, a new unnamed variable is created. In this way, we can regard these
 * variables as "temporary variables".
 *
 * No operations are supported for variables (that's a feature!) except
 * assignments between them. It is up to other classes (e.g. poly) to support
 * conversions from variables.
 *
 * Internally, such a mathematical variable is just coded by a single int.
 * This int can be written or read by the functions getnum and setnum; but
 * this is intended to be used only by other libraries, and not by the end
 * user.
 *
 * Note:
 * - Temporary variables can not be read by the istream >> operator. When
 *   written by the ostream << operator, they appear as #[n], where n is
 *   a unique number.
 *
 */

#ifndef __VAR_H__
#define __VAR_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "database.h"


int var_num(char,int);
extern int tmpvar;

class var {
	public:
		var () {number = next_tmp ();};
		var (int) {assert (0);}; // not supported
		var (char c, int i = 0) {number = var_num (c,i);}

		/*
		 * These direct access functions should only be used by other
		 * libraries that need to get and set variables fast
		 * 
		 */
		friend void setnum (var& v, int n);
		friend int getnum (const var& v);
		friend int var_num (char, int = 0); // get code number for c[i]

		// Comparison
		friend bool operator== (const var& v, const var& w);
		friend bool operator!= (const var& v, const var& w);

		// Input and output of variables
		friend ostream& operator<< (ostream&, const var&);
		friend istream& operator>> (istream&, var&);

		// Database functions
		friend void set_key (database& db, const var& v);
		friend void set_value (database& db, const var& v);
		friend bool get_value (database& db, var& v);
	private:
		int number;

		int next_tmp () { // get the next temporary variable
			return --tmpvar;
		};
};



// Direct access functions
void setnum (var& v, int n);
int getnum (const var& v);

// Comparison
bool operator== (const var& v, const var& w);
bool operator!= (const var& v, const var& w);

// Database functions
void set_key (database& db, const var& v);
void set_value (database& db, const var& v);
bool get_value (database& db, var& v);

#endif

